/*
 * Created by Aaron Lerch (2005)
 * Contact me at aaronlerch@yahoo.com
 */

using System;
using System.Text;

namespace Lerch.X10
{
    internal class CM17A : IX10Controller, IDisposable
    {
        private SerialPort serialPort;
        private string port;
		private object padlock = new object();

        #region Constants

        /// <summary>
        /// Delay time between bit toggling for CM17A (specified in Ticks)
        /// </summary>
        private const long DELAY_TIME = TimeSpan.TicksPerMillisecond / 2;
        /// <summary>
        /// First byte of standard CM17A header
        /// </summary>
        private const uint HEADER_1 = 0xD5;
        /// <summary>
        /// Second byte of standard CM17A header
        /// </summary>
        private const uint HEADER_2 = 0xAA;
        /// <summary>
        /// Standard CM17A footer
        /// </summary>
        private const uint FOOTER = 0xAD;
        /// <summary>
        /// CM17A brighten command byte
        /// </summary>
        private const uint BRIGHTEN_5_PERCENT = 0x88;
        /// <summary>
        /// CM17A dim command byte
        /// </summary>
        private const uint DIM_5_PERCENT = 0x98;
        /// <summary>
        /// First (of 2) byte adjustment for Unit codes 9-16
        /// </summary>
        private const uint FIRST_UPPER_UNIT_ID = 0x40;
        /// <summary>
        /// Adjustment for second data byte when turning off
        /// </summary>
        private const uint SECOND_UNIT_ID_OFF = 0x20;

        #endregion

        /// <summary>
        /// Binds to the CM17A device on the specified COM port
        /// </summary>
        /// <param name="port">COM port (e.g. COM1)</param>
        public CM17A(string port)
        {
            this.port = port;
            serialPort = new SerialPort(port);
            serialPort.Open();
            Init();
        }

        #region IX10Controller Members

        public void SendCommand(X10Command cmd)
        {
			if (Disposed) //verify in every method
			{
				throw new ObjectDisposedException("Object is already disposed");
			}

            switch (cmd.Command)
            {
                case Commands.ON:
                case Commands.OFF:
                    SetState(cmd.HouseCode, cmd.UnitCode, cmd.Command);
                    break;
                case Commands.BRIGHTEN:
                case Commands.DIM:
                    AdjustBrightness(cmd.HouseCode, cmd.UnitCode, cmd.Command, cmd.Level);
                    break;
            }
        }

        public void Initialize()
        {
			if (Disposed) //verify in every method
			{
				throw new ObjectDisposedException("Object is already disposed");
			}

            if (serialPort == null)
                serialPort = new SerialPort(port);

            if (!serialPort.IsOpen)
                serialPort.Open();

            Init();
        }

        public void Shutdown()
        {
			if (Disposed) //verify in every method
			{
				throw new ObjectDisposedException("Object is already disposed");
			}

            CloseSerialPort();
        }

        #endregion

        #region IDisposable Members

		// IDisposable implementation courtesy of IDesign (www.idesign.net)

		private bool m_disposed = false;

		protected bool Disposed
		{
			get
			{
				lock (padlock)
				{
					return m_disposed;
				}
			}
		}

		public void Dispose()
		{
			lock (padlock)
			{
				//Check to see if Dispose() has already been called
				if (m_disposed == false)
				{
					Cleanup();
					m_disposed = true;
					//Take yourself off the finalization queue 
					//to prevent finalization from executing a second time.
					GC.SuppressFinalize(this);
				}
			}
		}

		protected virtual void Cleanup()
		{
			CloseSerialPort();
		}

		//Destructor will run only if Dispose() is not called.
		//Do not provide destructors in types derived from this class.
		~CM17A()
		{
			Cleanup();
		}

        #endregion

        private void CloseSerialPort()
        {
            if ((serialPort != null) && (serialPort.IsOpen))
            {
                serialPort.Close();
                serialPort = null;
            }
        }

        #region Serial Port Control

        private void Init()
        {
            serialPort.RtsEnable = true;
            serialPort.DtrEnable = true;
        }

        private void Delay()
        {
            Delay(DELAY_TIME);
        }

        private void Delay(long ticks)
        {
            System.Threading.Thread.Sleep(TimeSpan.FromTicks(ticks));
        }

        private void Reset()
        {
            serialPort.RtsEnable = false;
            serialPort.DtrEnable = false;
        }

        #endregion

        #region Serial Port Communication

        /// <summary>
        /// Sends a single bit through the serial port
        /// </summary>
        /// <param name="state">True = 1, False = 0</param>
        private void SendBit(bool state)
        {
            if (!serialPort.DtrEnable || !serialPort.RtsEnable)
            {
                Init();
                Delay();
            }

            if (state)
            {
                serialPort.DtrEnable = false;  // Logical '1'
            }
            else
            {
                serialPort.RtsEnable = false;  // Logical '0'
            }

            Delay();
            Init();
            Delay();
        }

        /// <summary>
        /// Sends a byte to the CM17A
        /// </summary>
        /// <param name="pattern">Byte to send</param>
        private void SendByte(uint pattern)
        {
            SendBit((pattern & 0x80) == 0x80);
            SendBit((pattern & 0x40) == 0x40);
            SendBit((pattern & 0x20) == 0x20);
            SendBit((pattern & 0x10) == 0x10);
            SendBit((pattern & 0x08) == 0x08);
            SendBit((pattern & 0x04) == 0x04);
            SendBit((pattern & 0x02) == 0x02);
            SendBit((pattern & 0x01) == 0x01);
        }

        /// <summary>
        /// Sends the standard CM17A header sequence
        /// </summary>
        private void SendHeader()
        {
            SendByte(HEADER_1);
            SendByte(HEADER_2);
        }

        /// <summary>
        /// Sends the standard CM17A footer sequence
        /// </summary>
        private void SendFooter()
        {
            SendByte(FOOTER);
        }

        /// <summary>
        /// Sends a command (2 bytes).  Sends header, 2 bytes, and footer.
        /// </summary>
        private void SendCommand(uint firstByte, uint secondByte)
        {
            SendHeader();
            SendByte(firstByte);
            SendByte(secondByte);
            SendFooter();
        }

        #endregion

        /// <summary>
        /// Turns an X10 device on or off.
        /// </summary>
        /// <param name="house">HouseCode of device</param>
        /// <param name="unit">UnitCode of device</param>
        /// <param name="command">ON or OFF</param>
        private void SetState(HouseCode house, UnitCode unit, Commands command)
        {
            uint firstByte, secondByte;
            uint stateByte;

            if (command == Commands.ON)
                stateByte = 0;
            else
                stateByte = SECOND_UNIT_ID_OFF;

            firstByte = GetHouseByte(house) | (((int)unit >= 8) ? FIRST_UPPER_UNIT_ID : 0);
            secondByte = GetUnitByte(unit) | stateByte;
            SendCommand(firstByte, secondByte);
        }

        /// <summary>
        /// Sets the brightness level (if possible) of an X10 device
        /// </summary>
        /// <param name="house">HouseCode of device</param>
        /// <param name="unit">UnitCode of device</param>
        /// <param name="command">Command to execute</param>
        /// <param name="percent">Level to set device to</param>
        private void AdjustBrightness(HouseCode house, UnitCode unit, Commands command, int percent)
        {
            uint firstByte, secondByte;
            uint modeByte;
            int steps = (int)(percent / 5);

            if (command == Commands.BRIGHTEN)
                modeByte = BRIGHTEN_5_PERCENT;
            else
                modeByte = DIM_5_PERCENT;

            firstByte = GetHouseByte(house);
            secondByte = modeByte;
            while (steps-- > 0)
            {
                SendCommand(firstByte, secondByte);
            }
        }

        #region ENUM Translations

        /// <summary>
        /// Translates the common HouseCode enum into the specific bytes for this interface.
        /// </summary>
        private uint GetHouseByte(HouseCode house)
        {
            switch (house)
            {
                case HouseCode.A:
                    return 0x60;
                case HouseCode.B:
                    return 0x70;
                case HouseCode.C:
                    return 0x40;
                case HouseCode.D:
                    return 0x50;
                case HouseCode.E:
                    return 0x80;
                case HouseCode.F:
                    return 0x90;
                case HouseCode.G:
                    return 0xA0;
                case HouseCode.H:
                    return 0xB0;
                case HouseCode.I:
                    return 0xE0;
                case HouseCode.J:
                    return 0xF0;
                case HouseCode.K:
                    return 0xC0;
                case HouseCode.L:
                    return 0xD0;
                case HouseCode.M:
                    return 0x00;
                case HouseCode.N:
                    return 0x10;
                case HouseCode.O:
                    return 0x20;
                case HouseCode.P:
                    return 0x30;
            }

            throw new ArgumentException("Unrecognized HouseCode provided");
        }

        /// <summary>
        /// Translates the common UnitCode enum into the specific byte for this interface.
        /// </summary>
        private uint GetUnitByte(UnitCode unit)
        {
            switch (unit)
            {
                case UnitCode.U1:
                case UnitCode.U9:
                    return 0x00;
                case UnitCode.U2:
                case UnitCode.U10:
                    return 0x10;
                case UnitCode.U3:
                case UnitCode.U11:
                    return 0x08;
                case UnitCode.U4:
                case UnitCode.U12:
                    return 0x18;
                case UnitCode.U5:
                case UnitCode.U13:
                    return 0x40;
                case UnitCode.U6:
                case UnitCode.U14:
                    return 0x50;
                case UnitCode.U7:
                case UnitCode.U15:
                    return 0x48;
                case UnitCode.U8:
                case UnitCode.U16:
                    return 0x58;
            }

            throw new ArgumentException("Unrecognized UnitCode provided");
        }

        #endregion
    }
}
