/*
 * Created by Aaron Lerch (2005)
 * Contact me at aaronlerch@yahoo.com
 */

using System;
using System.Runtime;
using System.Runtime.InteropServices;

namespace Lerch.X10
{
	/// <summary>
	/// Summary description for SerialPort.
	/// </summary>
	public class SerialPort : IDisposable
	{
		private string m_strPort = String.Empty;
		private IntPtr hPort = IntPtr.Zero;
		private bool bOpen = false;

		private int _dtr = 0;
		private int _rts = 0;

		public bool RtsEnable
		{
			get
			{
				return (_rts == 1);
			}

			set
			{
				if (CheckOnline())
				{
					if (value)
					{
						if (Win32Com.EscapeCommFunction(hPort, Win32Com.SETRTS))
						{
							_rts = 1;
						}
						else
						{
							throw new Exception("Unexpected error setting RTS");
						}
					}
					else
					{
						if (Win32Com.EscapeCommFunction(hPort, Win32Com.CLRRTS))
						{
							_rts = 0;
						}
						else
						{
							throw new Exception("Unexpected error clearing RTS");
						}
					}
				}
				else
				{
					throw new Exception(m_strPort + " not open");
				}
			}
		}

		public bool DtrEnable
		{
			get
			{
				return (_dtr == 1);
			}

			set
			{
				if (CheckOnline())
				{
					if (value)
					{
						if (Win32Com.EscapeCommFunction(hPort, Win32Com.SETDTR))
						{
							_dtr = 1;
						}
						else
						{
							throw new Exception("Unexpected error setting DTR");
						}
					}
					else
					{
						if (Win32Com.EscapeCommFunction(hPort, Win32Com.CLRDTR))
						{
							_dtr = 0;
						}
						else
						{
							throw new Exception("Unexpected error clearing DTR");
						}
					}
				}
				else
				{
					throw new Exception(m_strPort + " not open");
				}
			}
		}

		public bool IsOpen
		{
			get
			{
				return ((hPort != (System.IntPtr)Win32Com.INVALID_HANDLE_VALUE) && (bOpen)) ? true : bOpen;
			}
		}

		public SerialPort(string port)
		{
			if ((port == null) || (port != null && port.Length == 0))
			{
				throw new ArgumentNullException("port", "Port must be specified");
			}
		
			m_strPort = port;
		}

		public void Open()
		{
			hPort = Win32Com.CreateFile(m_strPort, Win32Com.GENERIC_READ | Win32Com.GENERIC_WRITE, 0, IntPtr.Zero,
				Win32Com.OPEN_EXISTING, Win32Com.FILE_FLAG_OVERLAPPED, IntPtr.Zero);
			if (hPort == (IntPtr)Win32Com.INVALID_HANDLE_VALUE)
			{
				if (Marshal.GetLastWin32Error() == Win32Com.ERROR_ACCESS_DENIED)
				{
					Exception ex = new Exception(String.Format("Unable to open port {0}, access denied", m_strPort));
					throw ex;
				}
				else
				{
					// Try alternative name form if main one fails:
					hPort = Win32Com.CreateFile(AltName(m_strPort), Win32Com.GENERIC_READ | Win32Com.GENERIC_WRITE, 0, IntPtr.Zero,
						Win32Com.OPEN_EXISTING, Win32Com.FILE_FLAG_OVERLAPPED, IntPtr.Zero);
					if (hPort == (IntPtr)Win32Com.INVALID_HANDLE_VALUE)
					{
						if (Marshal.GetLastWin32Error() == Win32Com.ERROR_ACCESS_DENIED)
						{
							Exception ex = new Exception(String.Format("Unable to open port {0}, access denied", m_strPort));
							throw ex;
						}
						else
						{
							Exception ex = new Exception(String.Format("Failure to open port {0}", m_strPort));
							throw ex;
						}
					}
				}
			}

			bOpen = true;

			// Default DTR & RTS to off
			DtrEnable = false;
			RtsEnable = false;
		}

		/// <summary>
		/// Closes the COM port.
		/// </summary>
		public void Close()
		{
			if (bOpen)
			{
				Win32Com.CancelIo(hPort);
				Win32Com.CloseHandle(hPort);
				bOpen = false;
			}
		}

		private bool CheckOnline()
		{
			if ((hPort != (System.IntPtr)Win32Com.INVALID_HANDLE_VALUE) && (bOpen))
			{
				return true;
			}
			else
			{
				Open();
				return bOpen;
			}
		}

		/// <summary>
		/// Returns the alternative name of a com port i.e. \\.\COM1 for COM1:
		/// Some systems require this form for double or more digit com port numbers.
		/// </summary>
		/// <param name="s">Name in form COM1 or COM1:</param>
		/// <returns>Name in form \\.\COM1</returns>
		private string AltName(string s)
		{
			string r = s.Trim();
			if (s.EndsWith(":")) s = s.Substring(0, s.Length - 1);
			if (s.StartsWith(@"\")) return s;
			return @"\\.\" + s;
		}

		#region IDisposable Members

		// IDisposable implementation courtesy of IDesign (www.idesign.net)

		private bool m_disposed = false;

		protected bool Disposed
		{
			get
			{
				lock (this)
				{
					return m_disposed;
				}
			}
		}

		public void Dispose()
		{
			lock (this)
			{
				//Check to see if Dispose() has already been called
				if (m_disposed == false)
				{
					Cleanup();
					m_disposed = true;
					//Take yourself off the finalization queue 
					//to prevent finalization from executing a second time.
					GC.SuppressFinalize(this);
				}
			}
		}

		protected virtual void Cleanup()
		{
			Close();
		}

		//Destructor will run only if Dispose() is not called.
		//Do not provide destructors in types derived from this class.
		~SerialPort()
		{
			Cleanup();
		}

		#endregion
	}


	internal class Win32Com 
	{

		/// <summary>
		/// Opening Testing and Closing the Port Handle.
		/// </summary>
		[DllImport("kernel32.dll", SetLastError=true)]
		internal static extern IntPtr CreateFile(String lpFileName, UInt32 dwDesiredAccess, UInt32 dwShareMode,
			IntPtr lpSecurityAttributes, UInt32 dwCreationDisposition, UInt32 dwFlagsAndAttributes,
			IntPtr hTemplateFile);

		//Constants for errors:
		internal const UInt32 ERROR_FILE_NOT_FOUND = 2;
		internal const UInt32 ERROR_INVALID_NAME = 123;
		internal const UInt32 ERROR_ACCESS_DENIED = 5;
		internal const UInt32 ERROR_IO_PENDING = 997;
		internal const UInt32 ERROR_IO_INCOMPLETE = 996;

		//Constants for return value:
		internal const Int32 INVALID_HANDLE_VALUE = -1;

		//Constants for dwFlagsAndAttributes:
		internal const UInt32 FILE_FLAG_OVERLAPPED = 0x40000000;

		//Constants for dwCreationDisposition:
		internal const UInt32 OPEN_EXISTING = 3;

		//Constants for dwDesiredAccess:
		internal const UInt32 GENERIC_READ = 0x80000000;
		internal const UInt32 GENERIC_WRITE = 0x40000000;

		[DllImport("kernel32.dll")]
		internal static extern Boolean CloseHandle(IntPtr hObject);

		[DllImport("kernel32.dll")]
		internal static extern Boolean CancelIo(IntPtr hFile);

		/// <summary>
		/// Control port functions.
		/// </summary>
		[DllImport("kernel32.dll")]
		internal static extern Boolean EscapeCommFunction(IntPtr hFile, UInt32 dwFunc);

		// Constants for dwFunc:
		internal const UInt32 SETXOFF = 1;
		internal const UInt32 SETXON = 2;
		internal const UInt32 SETRTS = 3;
		internal const UInt32 CLRRTS = 4;
		internal const UInt32 SETDTR = 5;
		internal const UInt32 CLRDTR = 6;
		internal const UInt32 RESETDEV = 7;
		internal const UInt32 SETBREAK = 8;
		internal const UInt32 CLRBREAK = 9;
	}
}
