/*
 * Created by Aaron Lerch (2005)
 * Contact me at aaronlerch@yahoo.com
 */

using System;
using System.Collections;
using System.Threading;
using System.Text;

namespace Lerch.X10
{
    public class X10CommandProcessor : IDisposable
    {
        /// <summary>
        /// Number of seconds to delay between sending queued commands
        /// </summary>
        private const int DELAY_BETWEEN_COMMANDS = 1;

        /// <summary>
        /// The queue to hold X10 Commands to be sent.
        /// </summary>
        private Queue queue;

        /// <summary>
        /// The controller to accept queued commands.
        /// </summary>
        private IX10Controller x10Controller;

        /// <summary>
        /// Background thread for processing X10 commands
        /// </summary>
        private Thread processor = null;

        /// <summary>
        /// ManualResetEvent to synchronize background thread processing
        /// </summary>
        ManualResetEvent mre = new ManualResetEvent(false);

		/// <summary>
		/// Object used for obtaining synchronization lock
		/// </summary>
		private object padlock = new object();

        public X10CommandProcessor()
        {
            queue = new Queue();
        }

        public X10CommandProcessor(IX10Controller x10Controller) : this()
        {
            this.x10Controller = x10Controller;
        }

        /// <summary>
        /// The X10 Controller to send commands to.
        /// </summary>
        public IX10Controller X10Controller
        {
            get
            {
                return x10Controller;
            }
            set
            {
                x10Controller = value;
            }
        }

        /// <summary>
        /// Enqueues a command to be executed.
        /// </summary>
        public void SendCommand(X10Command cmd)
        {
			if (Disposed) //verify in every method
			{
				throw new ObjectDisposedException("Object is already disposed");
			}

            if ((cmd != null) && (x10Controller != null))
            {
                // If a new thread has not yet been created, 
                // create one to process queued commands
                if (processor == null)
                {
                    processor = new Thread(new ThreadStart(ProcessQueueCommands));
                    processor.Start();
                }
                if (processor.ThreadState == ThreadState.Stopped)
                    processor.Start();

                Object lockobj = new Object();
                lock (lockobj)
                {
                    queue.Enqueue(cmd);
                    mre.Set();
                }
            }
        }

        private void ProcessQueueCommands()
        {
            try
            {
                X10Command cmd = null;
                while (1 == 1)
                {
                    Object lockobj = new Object();
                    lock (lockobj)
                    {
                        if (queue.Count > 0)
                        {
                            cmd = queue.Dequeue() as X10Command;
                            if (queue.Count == 0)
                                mre.Reset();
                        }
                    }

                    if ((cmd != null) && (x10Controller != null))
                    {
                        x10Controller.SendCommand(cmd);
                        cmd = null;
                        // Sleep for at least 2ms between commands (if more than one)
                        Thread.Sleep(TimeSpan.FromSeconds(DELAY_BETWEEN_COMMANDS));
                    }

                    mre.WaitOne();
                }
            }
            catch (ThreadAbortException)
            {
                // Ignore, no action required
            }
        }

        #region IDisposable Members

		// IDisposable implementation courtesy of IDesign (www.idesign.net)

		private bool m_disposed = false;

		protected bool Disposed
		{
			get
			{
				lock (padlock)
				{
					return m_disposed;
				}
			}
		}

		public void Dispose()
		{
			lock (padlock)
			{
				//Check to see if Dispose() has already been called
				if (m_disposed == false)
				{
					Cleanup();
					m_disposed = true;
					//Take yourself off the finalization queue 
					//to prevent finalization from executing a second time.
					GC.SuppressFinalize(this);
				}
			}
		}

		protected virtual void Cleanup()
		{
			try
			{
				if (processor != null)
					processor.Abort();
			}
			catch (ThreadAbortException)
			{}
		}

		//Destructor will run only if Dispose() is not called.
		//Do not provide destructors in types derived from this class.
		~X10CommandProcessor()
		{
			Cleanup();
		}

        #endregion
    }
}
